#include <stdio.h>
#include <stdlib.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <io.h>
#include <fcntl.h>
#endif
#include <string.h>
#include <sys/param.h> 

#include "WODict.h"
#include "WOHTTPProtocol.h"
#include "WOAdaptorUtil.h"
#include "WOTransaction.h"
#include "WOURLCUtilities.h"
#include "WODirectoryListing.h"
#include "WORecording.h"

#define TIME_TEST 0 
#define LOAD_BALANCING 1 

#define QUOTE		'"'
	
#ifdef WIN32
__declspec(dllimport) const char* WOReadKeyFromConfiguration (char* keyName, char *buffer, int bufferSize);
#endif WIN32

extern char *WO_CGI_Headers[32];
extern char *WO_Req_Headers[32];

/* 
 * cgi_getenv(key) : get value from CGI environment
 */
static char *cgi_getenv(const char *envar)
{
    return getenv(envar);
}

/* die - print message and exit */
static void die(char *msg)
{
    fprintf(stdout,"Content-Type: text/html\n\n");
    fprintf(stdout,"<html><body><strong>%s</strong></body></html>\n",msg);
    fflush(stdout);
    exit(0);
}
/* 
 * get_http_method() find the HTTP method for this request
 */
static int get_http_method()
{
    char *method;
    method = cgi_getenv(CGI_REQUEST_METHOD);
    if (method) {    
	if (!strcmp(method,GET_METHOD)) {
	    return HTTP_GET_METHOD;
	} else if (!strcmp(method, POST_METHOD)) {
	    return HTTP_POST_METHOD;
	} else if (!strcmp(method, HEAD_METHOD)) {
	    return HTTP_HEAD_METHOD;
	} else if (!strcmp(method, PUT_METHOD)) { /* cern 0.9 */
	    return HTTP_PUT_METHOD;
	}
	return HTTP_UNKNOWN_METHOD;
    }
    return HTTP_NO_METHOD;
}
	
/* 
 * get_documentRoot() - return document root. 
 */
static char *get_documentRoot(char *pathInfo,char *buffer, int buflen)
{
    char *pathTranslated;
    
    if ((pathTranslated = cgi_getenv(CGI_PATH_TRANSLATED)) && pathInfo) {
	int len;
#ifdef WIN32
	char *p;
	while (p = strchr(pathTranslated,'\\')) {
	    p[0] = '/';
	}
#endif

	len = strlen(pathTranslated) - strlen(pathInfo);
	if(len >0 && (!strcmp(&pathTranslated[len],pathInfo))) { 
	    /* There is a possible security thing. If my doc root is
		/me/PrivateSecurePlace/importantThings/www, and then
		the CGI gives me
		/me/PrivateSecurePlace/importThings/www/foo/bar as 
		pathTranslated, and then /importThings/www/foo/bar as
		pathInfo, then they might be able to access things in
		/me/PrivateSecurePlace. But in real life (tm) that shouldnt
		really happen. If they hacked the server you got other
		problems */
	    if (len < buflen) {
		strncpy(buffer,pathTranslated,len);
		buffer[len] = '\0';
		return buffer;
	    }
	}
    } 
    buffer[0]='\0';
#ifdef WIN32
        {
            char documentRoot[WO_MAX_PATH+1];

            WOReadKeyFromConfiguration ("DOCUMENT_ROOT", documentRoot, WO_MAX_PATH);

            if (documentRoot && *documentRoot) {
                int docRootLen = strlen(documentRoot);
                if (docRootLen < buflen) {
                    strncpy (buffer, documentRoot, docRootLen);
                    buffer[docRootLen] = '\0';
                } else {
                    strncpy (buffer, documentRoot, buflen - 1);
                    buffer[buflen] = '\0';
                }
            }
        }
#endif

    return buffer;
}    

/* just return the input string, for either GET or POST methods. The
 * returned data must be freed
 */
static void * getContent()
{
    void *qs, *contentLength;
    int inputLen;
 
    if (get_http_method() == HTTP_POST_METHOD) {
	contentLength = cgi_getenv(CGI_CONTENT_LENGTH);
	if (!contentLength) return NULL;
	else {
	    inputLen = atol(contentLength);
	}
	qs =(void *)malloc(inputLen*sizeof(char)); 	
	if (fread(qs, sizeof(char),inputLen,stdin) != inputLen) {
	    /* I have to punt here. THE CGI doc says you have to have the right size or you suck. And if I dont get the right size, I am not sure the info I am getting is parsable. */
	    free(qs);
	    return NULL;
	} else {
	    return qs;
	}
    }
    return NULL;
}

void WOCGI()
{
    char *path_info, *script_name, *p;
    char *server_protocol = NULL;
    char *urlString = NULL;
    WOHTTPRequest *request = NULL;
    WOHTTPResponse *response = NULL;
    int appInstance=-1;
    WOURLComponents *url = NULL;
    char recordingDirPath[WO_MAX_PATH+1];
#ifdef WIN32
    WSADATA socketInitData;
#endif

/* intermediate stuff */
    char *appName = NULL;
    char *hostName = NULL;
    char documentRoot[WO_MAX_PATH+1];
    int	method;			/* http method */
    int len;
    char **cgi_header, **req_header;
#ifdef WIN32
    WSAStartup(0x0101, &socketInitData);
#endif

#ifdef WIN32
    _setmode(_fileno(stdout), _O_BINARY);
    _setmode(_fileno(stdin), _O_BINARY);
#endif

    method = get_http_method();
    if ((method == HTTP_PUT_METHOD) || (method == HTTP_HEAD_METHOD) || (method == HTTP_UNKNOWN_METHOD)) {
	die("Unknown HTTP method, or HTTP PUT or HEAD method not supported.");
    }
    if (method == HTTP_NO_METHOD) {
	fprintf(stderr,"Exiting : WebObjects is a cgi program that should only be started by an http server.\n");
	exit(0);
    }

    if (!(path_info = cgi_getenv(CGI_PATH_INFO))) {
/* Bad URL.
at this stage, if there is no config file, we give a listing */
	if (!configFileExists()) {
            url = (WOURLComponents *)malloc(sizeof(WOURLComponents));
            WOParseURL(url, cgi_getenv(CGI_SCRIPT_NAME));
            get_documentRoot("",documentRoot,sizeof(documentRoot));
            response = WODirectoryListing("", documentRoot, url);            
	} else {
            die("No application name and other arguments in URL, or path_info too long");
	}
    }

    if (!response) {	
#ifdef WIN32
        while(p = strchr(path_info,'\\')) {
            p[0]='/';
        }
#endif

        get_documentRoot(path_info,documentRoot,sizeof(documentRoot));

        if (!(server_protocol = cgi_getenv(CGI_SERVER_PROTOCOL))) {
            die("Cannot determine server protocol from CGI");
        }

        if (!(script_name = cgi_getenv(CGI_SCRIPT_NAME))) {
            die("Cannot determine script name from CGI");
        }
#ifdef WIN32
        while(p = strchr(script_name,'\\')) {
            p[0]='/';
        }
#endif

        logWebObjects(LOGCGI_INFO,"-- WOCGI: script_name = %s\n\tserver_protocol = %s\n\tpath_info = %s",script_name, server_protocol, path_info);

        len = strlen(script_name)+strlen(path_info);
        urlString = malloc((len+1)*sizeof(char));
        sprintf(urlString,"%s%s",script_name,path_info);

/* Do not worry, this will only do something in the case of the WebObjects-Recording adaptor */
        if (getRecordingDir(cgi_getenv(CGI_QUERY_STRING), recordingDirPath, sizeof(recordingDirPath)) < 0) {
            response = WOErrorResponseWithFile(documentRoot, ERROR_RESPONSE_NO_RECORDING_FILE);
        }


        url = (WOURLComponents *)malloc(sizeof(WOURLComponents));
        WOParseURL(url, urlString);
        switch(WOCheckURL(url)) {
            case WOURLOK:
                break;
            case WOURLInvalidPrefix:
                die("Invalid prefix in URL");
                break;
            case WOURLInvalidApplicationName:
                /* If there is no config file, we give a listing */
                if (!configFileExists()) {
                    response = WODirectoryListing("", documentRoot, url);
                    break;
                }
                die("Invalid application name in URL");
                break;
            case WOURLInvalidApplicationNumber:
                die("Invalid application number in URL");
                break;
            case WOURLInvalidApplicationHost:
                die("Invalid application host name in URL");
                break;
            case WOURLInvalidPageName:
                die("Invalid page name in URL");
                break;
            case WOURLInvalidSessionID:
                die("Invalid session ID in URL");
                break;
            case WOURLInvalidContextID:
                die("Invalid context ID in URL");
                break;
            case WOURLInvalidSenderID:
                die("Invalid sender ID in URL");
                break;
            case WOURLInvalidQueryString:
                die("Invalid query string in URL");
                break;
            default:
                die("Invalid URL");
                break;
        }
    }

    if (!response) {
        if (url->applicationName.length) {
                appName = malloc((url->applicationName.length+1)*sizeof(char));
                strncpy(appName,url->applicationName.start,url->applicationName.length);
                appName[url->applicationName.length] = '\0';
                }
        if (url->applicationHost.length) {
                hostName = malloc((url->applicationHost.length+1)*sizeof(char));
                strncpy(hostName,url->applicationHost.start,url->applicationHost.length);
                hostName[url->applicationHost.length] = '\0';
                }
        if (url->applicationNumber.length) {
                char *appInstanceString;
                appInstanceString = malloc((url->applicationNumber.length+1)*sizeof(char));
                strncpy(appInstanceString,url->applicationNumber.start,url->applicationNumber.length);
                appInstanceString[url->applicationNumber.length] = '\0';
                appInstance = atol(appInstanceString);
                free(appInstanceString);
                }

        if (!appName) {
            die("Cannot obtain the name of the application being requested, or the application name is too long");
        }

        logWebObjects(LOGCGI_INFO,"-- WOCGI: \tAppName = %s\n\tInstance = %d\n\tHostName = %s\n\tdocumentRoot = %s",appName ? appName : "NONE", appInstance, hostName ? hostName : "NONE", (documentRoot && *documentRoot) ? documentRoot : "NONE");

/* 
* Build the request
*/
        request = (WOHTTPRequest *) malloc(sizeof(WOHTTPRequest));


        request->headers = dict_create(30);

/* Adds the private document root header. It will be removed at the adaptor level in the
* application adaptor.
*/

        if (documentRoot && *documentRoot) {
            dict_nvinsert(DOCUMENT_ROOT,documentRoot,request->headers);
        }

/* fill in all the HTTP related information */
        cgi_header = WO_CGI_Headers;
        req_header = WO_Req_Headers;

        while (*cgi_header && *req_header) {
                if ((p = cgi_getenv(*cgi_header))) {
                    dict_nvinsert(*req_header,p,request->headers);
                }
            cgi_header++;
            req_header++;
        }

/* Build the request line */

        if (method == HTTP_GET_METHOD) {
            char *qs = cgi_getenv(CGI_QUERY_STRING);
            url->queryString.start = qs;
            url->queryString.length = qs ? strlen(qs) : 0;
            request->content = NULL;
        } else {
            request->content = getContent();
        }


/* just to say that we are not using netscape api here */
        request->isNSAPI = 0;
/* just to say that we are not using ISAPI here */
        request->isISAPI = 0;
        response = TransactionWithAutoStart(appName,appInstance,hostName, documentRoot, 300, request, url, server_protocol, recordingDirPath);
    }
    
    if (!response) {
        die("Did not receive any response from application. It is possible that the application does not exist, or that the requested url is incorrect.");
    }

    {
	char *version, *msg, *newline, *length;
	int stat,i;
    
	len = strlen(response->statusLine);
	version = malloc((len+1)*sizeof(char));
	msg = malloc((len+1)*sizeof(char));
	newline = malloc((len*2+1)*sizeof(char));
	sscanf(response->statusLine,"%s %d %s",version,&stat,msg);
	sprintf(newline,"%s: %d %s\n",STATUS,stat,msg);
	fputs(newline,stdout);
	for (i=0;i < response->headers->hsize;i++) { /* headers */
	    struct dict_entry *dictEntry;

	    dictEntry = response->headers->ht[i];
	    while (dictEntry) {
		dict_param *dictParam;

		dictParam = dictEntry->param;
		if (dictParam && dictParam->name && dictParam->value) {
		    fputs(dictParam->name, stdout);
		    fputs(": ", stdout);
		    fputs(dictParam->value, stdout);
		    fputs("\n", stdout);
    
		}
		dictEntry = dictEntry->next;
		if (!dictEntry) break;
	    }
	}
	fputs("\n",stdout);
	if ((length = dict_findval(CONTENT_LENGTH,response->headers))) {
	    int contentLength = atoi(length);
#ifdef WIN32
	    int pointer = 0;
	    fflush(stdout);  /* flush it */
	    while (contentLength) {	/* Write the content */
		int len = _write(_fileno (stdout), response->content + pointer, contentLength);

		if (len == -1) {
		    logWebObjects(LOGCGI_INFO,"-- WOCGI: Writing contents failed after %s bytes", pointer);
		    break;
		}

		contentLength -= len;
		pointer += len;
	    }
#else
	    if (contentLength) {	/* Write the content */
		fwrite(response->content, contentLength,1,stdout);
	    }
#endif
	}
        if (version) {
            free(version); 
        }
        if (msg) {
            free(msg);
        }
        if (newline) {
            free(newline);
        }                     
    }
    fflush(stdout);  /* flush it */

/* no longer needed */
    free(urlString);
    free(response);
    if (appName) {
        free(appName);
    }
    if (hostName) {
        free(hostName);
    }
    if (url) {
        free(url);
    }
    if (request) {
        free(request);
    }
#ifdef WIN32
    WSACleanup();
#endif

    exit(0);
}
