/* This file contains the default behavior of the Server Adaptor. For the moment this means only
* the load balancing function.
*/
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#ifndef WIN32
#include <unistd.h>
#else
#include <sys/param.h>
#include <process.h>
#include <winsock.h>
#endif

#if (defined(NeXT_PDO) && defined(__svr4__))
#include <sys/param.h>
#endif

#ifdef NeXT_PDO
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef NeXT
#include <libc.h>
#endif

#include "WOServerAdaptor.h"
#include "WOAdaptorUtil.h"
#include "WOConfigFile.h"
#include "WOSocket.h"

#if defined(WIN32)
#include <sys/timeb.h>
#include <time.h>

#if ! defined(_WINSOCKAPI_)    // winsock.h defines timeval
struct timeval {
    long tv_sec;
    long tv_usec;
};
#endif // _WINSOCKAPI_

struct timezone {
    int tz_minuteswest;
    int tz_dsttime;
};

extern int gettimeofday(struct timeval *tp, struct timezone *tzp);

int gettimeofday(struct timeval *tp, struct timezone *tzp)
{
    struct _timeb tb;

    _ftime(&tb);
    if (tp) {
		tp->tv_sec = tb.time;
		tp->tv_usec = tb.millitm * 1000;
    }
    if (tzp) {
        tzp->tz_minuteswest = tb.timezone;
		tzp->tz_dsttime = _daylight;
    }

    return 0;
}

#else /* WIN32 */
#include <sys/time.h>
#include <sys/types.h>
#include <sys/signal.h>
#if (NeXT_PDO && __svr4__)
extern int gettimeofday(struct timeval *, void *);
extern int gethostname(char *, int);
#endif
#endif /* WIN32 */

#define GETTOD(ntp)	gettimeofday(ntp, NULL);

#ifndef MIN
#define MIN(a,b) (((a)<(b))?(a):(b))
#endif

/* allocates the appArray and fill it with NULL pointers
 * the actual allocation of each string is done as needed
 */

char **allocAppArray() {
    char **appArray;
    int i;
    
    appArray = malloc(MAX_NUMBER_OF_APP*sizeof(char *));
    for (i = 0; i < MAX_NUMBER_OF_APP; i++) {
        appArray[i] = NULL;
    }
    return appArray;
}

/* This will free the appArray only up to the last non-NULL string */

void freeAppArray(char **appArray) {
    int i;
    
    for (i = 0; i < MAX_NUMBER_OF_APP; i++) {
        if (appArray && appArray[i]) {
            free(appArray[i]);
        } else {
            break;
        }
    }
    if (appArray) free(appArray);
}

/* LOAD BALANCING enables to choose an application on a host from a list given on a configuration file, just knowing the name of the application (example : {appName = "WebApp"; hostName = NULL}). Here two things can happen :

Using Sockets, the conf file will be read and something like
{appName = "WebApp"; hostName = "myHost"; appPort = 2000;instance = 1} will be generated from a conf file with the lines :
WebApp:1@myHost 3000
WebApp:2@myHost 2000
...

The syntax is <ApplicationName>:<InstanceNumber>[@<HostName>] <portNumber>

When the host is not specified then gethostname() is used to set it to the server host.

The host and the port number of the WebApp are still picked up randomly.
*/

int WOServerAdaptorGetRandomHostNameAndPort(const char *appName, int *instance, char **hostName, int *appPort, int *sockDes, int timeout, const char *configFile)
{
    FILE *fp;
    char app[MAX_LENGTH_CONF_LINE];
    char **appArray;
    char name[MAX_LENGTH_CONF_LINE];
    int portArray[MAX_NUMBER_OF_APP];

    int nApps = 0;
    
    *sockDes = -1;

    if (!(appName && *appName)) {
	logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: Application name is NULL or empty while load balancing\n");
	return LOAD_BALANCING_FAILURE;
    }

    logWebObjects(LOGCGI_INFO,"-- WOServerAdaptor: Load Balancing for %s",appName);

    if (!(fp = WOOpenConfigFile(configFile))) {
        return LOAD_BALANCING_NO_CONFIG;
    }
	

    appArray = allocAppArray();
    while (getLineFromFilePointer(app, MAX_LENGTH_CONF_LINE-1,fp))
	{
	if (app[0] != '\0') {
	    char *pos = strchr(app,':');
	    if (!pos) {
		logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: line without application number in conf file :%s", CONFIG_FILE);
		fclose(fp);
                freeAppArray(appArray);
		return LOAD_BALANCING_FAILURE;
            }
   
	    strncpy(name,app,pos-app);
	    name[pos-app]='\0';
	    if (!strcmp(name, appName))
		{
                /* First we allocate a new application info line in appArray */
                appArray[nApps] = malloc(MAX_LENGTH_CONF_LINE*sizeof(char));

                /* Now we get the application info and the port */               
                if (sscanf(pos, "%s %d",appArray[nApps],&portArray[nApps])!= 2) {
		    logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: Invalid format in configuration file %s : \"%s\" occured while parsing %s\n",CONFIG_FILE,strerror(errno),app);
		    fclose(fp);
                    freeAppArray(appArray);
		    return LOAD_BALANCING_FAILURE;
		}
		nApps++;
	    }
	}
    }

    if (ferror(fp)) {
	logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: \"%s\" occured while reading the configuration file %s\n",strerror(errno),CONFIG_FILE);
	fclose(fp);
        freeAppArray(appArray);
        return LOAD_BALANCING_FAILURE;
    }
    fclose(fp);

    if (nApps)
	{
	
	while (*sockDes < 0) {
	    int process = 0;
	    char *appChosen;
	    char *newHostName;
	    struct timeval tp;
	    tp.tv_usec = 0;
	    GETTOD(&tp);

/* On NT, the time in milliseconds is always even. I divide it by 2. 
 * It's a long, anyway.
 */

#ifdef WIN32
	    tp.tv_usec >>= 1;
#endif

#if defined(NeXT)
	srandom(tp.tv_usec);
	process = random() % nApps;
#else
        srand(tp.tv_usec);
        process = rand() % nApps;
#endif

/* because of the : we have to add 1 */
	    appChosen = appArray[process]+1;
    
    /* read the instance number */
	    *instance = atoi(appChosen);
    /* now read the host */
	    newHostName = strchr(appChosen,'@');
    
	    if (newHostName) {
		char *string = newHostName+1;
		int hostNameLength;
		hostNameLength = appChosen +strlen(appChosen) - newHostName;
		newHostName = malloc((hostNameLength + 1)*sizeof(char));
		strncpy(newHostName,string,hostNameLength);
		newHostName[hostNameLength] = '\0';
	    }
    /* else get host name from gethostname() */
	    else {
		char host[256];
		if (gethostname(host,sizeof(host)) == -1) {
		    logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: \"%s\" occured getting host name with gethostname()\n",strerror(errno));
                    freeAppArray(appArray);
                    return LOAD_BALANCING_FAILURE;
                }
		if (host && *host) {
		    newHostName = malloc((strlen(host)+1)*sizeof(char));
		    strcpy(newHostName,host);
		}
		else {
		    logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: gethostname() returns NULL host name\n");
                    freeAppArray(appArray);
                    return LOAD_BALANCING_FAILURE;
                }
	    }

/* Bug : here there is a leak but it is too late to fix it ...
 	    if (*hostName) {
	    	free(*hostName);
	    }
 */	    
	    *hostName = newHostName;
	    *appPort = portArray[process];
	    logWebObjects(LOGCGI_INFO,"-- WOServerAdaptor : connecting to app %s on host %s with port %d", appName, newHostName,*appPort);
	    *sockDes = WOSocketConnectTo(newHostName, *appPort, timeout);
	    if (*sockDes < 0) {
		nApps--;
		if (!nApps) {
		    logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor : Cannot connect to any application %s listed in configuration file.", appName);
                    freeAppArray(appArray);
                    return LOAD_BALANCING_FAILURE;
		}

/* Fix for bug #63864. If we could not connect to this chosen application, we pick an other.
 * But first we remove it from the array by replacing it with the last one
 * if it is not already the last one, in which case we do nothing.
 */
 		if (process != nApps) {
		    strncpy(appArray[process], appArray[nApps], MAX_LENGTH_CONF_LINE-1);
		    appArray[process][MAX_LENGTH_CONF_LINE-1] = '\0';
		    portArray[process] = portArray[nApps];
		    
		}				
	    }
	}
   }
   else {
	logWebObjects(LOGCGI_WARNING,"-- WOServerAdaptor: Application %s not listed in configuration file %s\n",appName,CONFIG_FILE);
       freeAppArray(appArray);
       return LOAD_BALANCING_APPLICATION_NOT_IN_CONFIG;
   }

    freeAppArray(appArray);

    return LOAD_BALANCING_SUCCESS;
}
